@@ -5,11 +5,13 @@ Created on Wed Jan 26 00:33:29 2022
 @author: Andrew Freiburger
 """
 from scipy.constants import hour
-from Bio.Blast import NCBIWWW
+from Bio.Blast import NCBIWWW, NCBIXML
+from warnings import warn
 from chemw import Proteins
 #from pprint import pprint
 from math import ceil
 from glob import glob
+import requests, io
 import datetime
 import json, os, re
 
@@ -62,6 +64,7 @@ class Codons():
     def __init__(self,
                  sequence: str = None,  # the genetic sequence can be optionally provided, for easy use in the other functions.
                  codons_table: str = 'standard', # the translation table for codons to amino acids
+                 start_codons: list = ["ATG", "AUG"], # the list of start codons that are accepted by the code
                  amino_acids_form: str = 'one_letter', # selects the scale of amino acid nomenclature
                  hyphenated: bool = None, # selects whether the printed protein will be hyphenated between the protein residues
                  verbose: bool = False,
@@ -69,13 +72,13 @@ class Codons():
                  ):
         self.verbose = verbose
         self.printing = printing
-        self.proteins = {}
+        self.genes = {}
         self.transcribed_sequence = None
-        self.protein_blast_results = None
+        self.protein_blast_results = []
         self.nucleotide_blast_results = None
         self.gene_fasta = None
         self.protein_fasta = None
-        self.protein_mass = Proteins()
+        self.protein_mass = Proteins(printing = False)
         
         # define the simulation paths
         self.paths = {}
@@ -85,6 +88,7 @@ class Codons():
         
         self.parameters = {}
         self.parameters['residue_delimiter'] = '-' 
+        self.parameters['start_codons'] = start_codons
         
         # refine the sequence into the FASTA format
         self.sequence = sequence
@@ -117,9 +121,9 @@ class Codons():
         for cd in self.changed_codons:
             self.codons_table[cd] = self.changed_codons[cd]
     
-    def _read_fasta(self,
+    def read_fasta(self,
                     fasta_path: str = None,  # the path to the fasta file
-                    fasta_link: str = None,  # the path to the fasta file
+                    fasta_link: str = None,  # the url to the fasta file
                     ):
         # import and parse fasta-formatted files        
         if fasta_path is not None:
@@ -129,14 +133,26 @@ class Codons():
             sequence = requests.get(fasta_link).content
             self.fasta_lines = io.StringIO(sequence.decode('utf-8')).readlines()
     
-        sequence = ''
+        sequences = []
+        descriptions = []
+        seq = ''
+        first = True
         for line in self.fasta_lines:
-            if not re.search('^>', line):
-                line = line.rstrip()
-                sequence += line
+            if not re.search('(^>)', line):
+                seq += line.rstrip()
             else:
-                description = line
-        return sequence, description
+                if first:
+                    first = False
+                    continue
+                descriptions.append(line)
+                sequences.append(seq)
+                seq = ''
+        if sequences == []:
+            sequences.append(seq)
+            seq = re.sub('([^A-Z])', '', seq, flags = re.IGNORECASE)
+            descriptions.append(' - '.join(['Sequence', f'{len(seq)}_monomers']))
+                
+        return sequences, descriptions, self.make_fasta(sequences, descriptions)
     
     def _paths(self, 
                export_name = None, 
@@ -156,8 +172,8 @@ class Codons():
                 tag = 'BLAST'
             else:
                 tag = 'BLASTp'
-        elif self.proteins != {}:
-            tag = f'{len(self.proteins)}_proteins'
+        elif self.genes != {}:
+            tag = f'{len(self.genes)}_proteins'
         elif self.transcribed_sequence:
             tag = f'{self.transcription}'
         if export_name is None:
@@ -186,16 +202,29 @@ class Codons():
         return export_path
             
     def make_fasta(self,
-                   sequence: str,  # the genetic nucleotide sequence
-                   description: str = 'sequence', # the description of the genetic or protein sequence
+                   sequences: str,  # the genetic nucleotide sequence
+                   descriptions: str = 'sequence', # the description of the genetic or protein sequence
                    export_path: str = None
                    ):
-        if sequence is None:
+        if sequences is None:
             return None
-        if not re.search('\*', sequence):
-            sequence += '*'
             
-        fasta_file = '\n'.join([f'>{description}',sequence])
+        if type(sequences) is list:
+            fasta_file = []
+            for index, sequence in enumerate(sequences):
+                description = descriptions[index]
+                if not re.search('\*', sequence):
+                    sequence += '*'
+                fasta = '\n'.join(['>'+description, sequence])
+                fasta_file.append(fasta)
+            fasta_file = '\n'.join(fasta_file)
+        elif type(sequences) is str:
+            if not re.search('\*', sequences):
+                sequences += '*'
+            fasta_file = '\n'.join(['>'+descriptions, sequences])
+        else:
+            ValueError(f'The sequence type {type(sequences)} is not supported by the function.')
+        
         if export_path is not None:
             with open(export_path, 'w') as out:
                 out.write(fasta_file)
@@ -210,14 +239,20 @@ class Codons():
                    ):
         if sequence:
             self.sequence = sequence
-            self.gene_fasta = self.make_fasta(self.sequence, ' - '.join(['Genetic_sequence', f'{len(self.sequence)}_bps']))
+            base_pairs = len(re.sub("([^atcgu])", "", self.sequence, flags = re.IGNORECASE))
+            unit = 'bp'
+            if base_pairs > 1e4:
+                unit = 'kbp'
+            self.gene_fasta = self.make_fasta(self.sequence, ' - '.join(['Genetic_sequence', f'{base_pairs}_{unit}']))
         elif fasta_path:
-            self.sequence, description = self._read_fasta(fasta_path)
+            sequences, descriptions, self.gene_fasta = self.read_fasta(fasta_path)
+            self.sequence = sequences[0]
         elif fasta_link:
-            self.sequence, description = self._read_fasta(fasta_link = fasta_link)
+            sequences, descriptions, self.gene_fasta = self.read_fasta(fasta_link = fasta_link)
+            self.sequence = sequences[0]
             
         # determine the capitalization of the sequence
-        for ch in sequence:
+        for ch in self.sequence:
             if re.search('[a-zA-Z]',ch): 
                 upper_case = ch.isupper()
                 break
@@ -241,11 +276,18 @@ class Codons():
             description = f'>Transcribed sequence from {self.transcription}'
         self.transcribed_fasta = self.make_fasta(self.transcribed_sequence, description)
         return self.transcribed_sequence
-        
+    
+    def _find_start(self,i):
+        starts = '(?=' + '|'.join(self.parameters['start_codons']) + ')'            
+        codon = ''
+        while not re.search(starts, codon, flags=re.IGNORECASE) and i < len(self.sequence):
+            codon = self.sequence[i:i+3]
+            i += 1
+        i -= 1
+        return i
         
     def translate(self,
                  sequence: str = None, # the genomic code as a string
-                 description: str = None,  # a description of the sequence
                  fasta_path: str = None,  # the path to the fasta file
                  fasta_link: str = None,  # the path to the fasta file
                  ):
@@ -253,55 +295,73 @@ class Codons():
             self.sequence = sequence
             self.gene_fasta = self.make_fasta(self.sequence, ' - '.join(['Genetic_sequence', f'{len(self.sequence)}_bps']))
         elif fasta_path:
-            self.sequence, description = self._read_fasta(fasta_path)
+            sequences, descriptions, self.gene_fasta = self.read_fasta(fasta_path)
+            self.sequence = sequences[0]
         elif fasta_link:
-            self.sequence, description = self._read_fasta(fasta_link = fasta_link)
+            sequences, descriptiosn, self.gene_fasta = self.read_fasta(fasta_link = fasta_link)
+            print(sequences)
+            self.sequence = sequences[0]
             
         self.protein_fasta = []
         self.missed_codons = []
-        codon = ''
         amino_acids = None
-        for nuc in self.sequence:
-            if re.search('[atucg]', nuc, flags = re.IGNORECASE):
+        self.sequence = re.sub('([^atucg])', '', self.sequence, flags = re.IGNORECASE)
+        
+        i = 0
+        while i < len(self.sequence):
+            i = self._find_start(i)
+            if i > len(self.sequence)-10:
+                break
+            codons = []
+            amino_acids = []
+            
+            codon = ''
+            gene = ''
+            old_i = i
+            for nuc in self.sequence[old_i:]:
+                i += 1
+                gene += nuc
                 codon += nuc
                 if len(codon) == 3:
+                    codons.append(codon)
                     if codon not in self.codons_table:
                         self.missed_codons.append(codon)
                     else:
                         amino_acid = self.codons_table[codon]
-                        if amino_acid == 'start':
-                            amino_acids = []
-                        elif amino_acid == 'stop' and type(amino_acids) is list:
-                            if len(amino_acids) >= 1:
-                                protein = self.parameters['residue_delimiter'].join(amino_acids)
-                                mass = self.protein_mass.mass(protein)
-                                
-                                self.proteins[protein] = mass
-                                if not description:
-                                    description = ' - '.join(['Protein', f'{len(protein)}_residues', f'{mass}_amu'])
-                                fasta_file = self.make_fasta(protein, description)
-                                self.protein_fasta.append(fasta_file)
-                                amino_acids = None
+                        if amino_acid == 'stop' and type(amino_acids) is list:
+                            protein = self.parameters['residue_delimiter'].join(amino_acids)
+                            mass = self.protein_mass.mass(protein)
+                            
+                            self.genes[gene] = {}
+                            self.genes[gene]['protein'] = {
+                                    'sequence': protein,
+                                    'mass': mass
+                            }
+                            
+                            description = ' - '.join(['Protein', f'{len(amino_acids)}_residues', f'{mass}_amu'])
+                            fasta_file = self.make_fasta(protein, description)
+                            self.protein_fasta.append(fasta_file)
+                            amino_acids = None
+                            break
                         else:
-                            if type(amino_acids) is list and re.search('[a-z]+',amino_acid, flags = re.IGNORECASE):
+                            if type(amino_acids) is list and re.search('[a-z]+', amino_acid, flags = re.IGNORECASE):
                                 amino_acids.append(amino_acid)
                                 if self.verbose:
                                     print(codon, '\t', amino_acid)
-                    
                     codon = ''
-                    
+            self.genes[gene]['codons'] = codons
+                
         self.protein_fasta = '\n'.join(self.protein_fasta)
-        if self.printing:
-           print(self.protein_fasta)
-           if self.missed_codons != []:
-               print(f'The {self.missed_codons} codons were not captured by the employed codons table.')
+        if self.printing:       
+            print(self.protein_fasta)
+            if self.missed_codons != []:
+                print(f'The {self.missed_codons} codons were not captured by the employed codons table.')
             
-        return self.proteins
     
     def blast_protein(self,  # https://blast.ncbi.nlm.nih.gov/Blast.cgi?PROGRAM=blastp&PAGE_TYPE=BlastSearch&LINK_LOC=blasthome
                       sequence: str = None,
                       database: str = 'nr', # the blastp database that will be searched with the collected FASTA sequences
-                      description: str = '',  # a description of the sequence
+                      description: str = 'Protein sequence description',  
                       fasta_path: str = None, # The path to a fasta file
                       fasta_link: str = None,  # the path to the fasta file
                       export_name = None, 
@@ -309,31 +369,38 @@ class Codons():
                       ):
         if sequence:
             self.sequence = sequence 
+            sequences = [sequence]
             self.protein_fasta = self.make_fasta(sequence, description) 
         elif fasta_path:
-            with open(fasta_path) as input:
-                self.protein_fasta = input.read() 
+            sequences, descriptions, self.protein_fasta = self.read_fasta(fasta_path = fasta_path)
         elif fasta_link:
-            sequence = requests.get(fasta_link).content
-            self.protein_fasta = io.StringIO(sequence.decode('utf-8')).read()
+            sequences, descriptions, self.protein_fasta = self.read_fasta(fasta_link = fasta_link)
             
         # estimate the completion time
         estimated_time = datetime.datetime.now()+datetime.timedelta(seconds = len(self.protein_fasta)/2)    # approximately 1/2 second per amino acid   
         print(f'The database search for the parameterized protein(s) will complete circa {estimated_time}.')
         
         # acquire the BLAST results
-        self.protein_blast_results = NCBIWWW.qblast('blastp', database, self.protein_fasta)
+        for index, sequence in enumerate(sequences):
+            short_query = False
+            if len(sequence) < 30:
+                short_query = True
+            self.protein_blast_results.append(NCBIWWW.qblast('blastp', database, sequence, short_query = short_query))
+            print(f' \rCompleted searches: {index+1}/{len(sequences)}\t{datetime.datetime.now()}', end='')
         
         # export the content
+        self.protein_blast_results = '\n\n'.join(self.protein_blast_results)
         self.export(export_name, export_directory) 
         self.paths['protein_blast_results'] = os.path.join(self.export_path, 'protein_blast_results.xml')
         with open(self.paths['protein_blast_results'], 'w') as protein_data:
             protein_data.write(self.protein_blast_results.read())
+            
+        return NCBIXML.read(self.protein_blast_results)
         
     def blast_nucleotide(self,
                          sequence: str = None,
                          database: str = 'nt',
-                         description: str = '',  # a description of the sequence
+                         description: str = 'Genetic sequence description',  # a description of the sequence
                          fasta_path: str = None, # The path to a fasta file
                          fasta_link: str = None,  # the path to the fasta file
                          export_name = None, 
@@ -342,14 +409,13 @@ class Codons():
         self.nucleotide_blast_results = []
         if sequence:
             self.sequence = sequence 
+            sequences = [sequence]
             self.gene_fasta = self.make_fasta(sequence, description) 
         elif fasta_path:
-            with open(fasta_path) as input:
-                self.gene_fasta = input.read() 
+            sequences, descriptions, self.gene_fasta = self.read_fasta(fasta_path = fasta_path)
         elif fasta_link:
-            sequence = requests.get(fasta_link).content
-            self.gene_fasta = io.StringIO(sequence.decode('utf-8')).read()
-            
+            sequences, descriptions, self.gene_fasta = self.read_fasta(fasta_link = fasta_link)
+                        
         # estimate the completion time
         estimated_length = len(self.gene_fasta)/2
         estimated_time = datetime.datetime.now()+datetime.timedelta(seconds = estimated_length)    # approximately 1 second per nucleic acid
@@ -361,21 +427,22 @@ class Codons():
         self.export(export_name, export_directory)
         self.paths['nucleotide_blast_results'] = [os.path.join(self.export_path, 'nucleotide_blast_results.xml')]
         
-        section_size = 2000
-        sections = ceil(len(self.gene_fasta)/section_size)
-        sequence_sections = [self.gene_fasta[i*section_size:(i+1)*section_size] for i in range(0, sections)]
-        for sequence in sequence_sections:
-            nucleotide_blast_result = NCBIWWW.qblast('blastn', database, self.gene_fasta)
-            result = nucleotide_blast_result.read()
-            self.nucleotide_blast_results.append(result)
-            
-            # export this modular portion
-            export_path = self._paths('nucleotide_blast_results.xml', self.export_path)
-            self.paths['nucleotide_blast_results'].append(export_path)
-            with open(self.paths['nucleotide_blast_results'][-1], 'w') as nucleotide_data:
-                nucleotide_data.write(result)
-            
-            print(f'Section {sequence_sections.index(sequence)+1}/{sections} is completed: {datetime.datetime.now()}')
+        section_size = 4000
+        for sequence in sequences:
+            sections = ceil(len(sequence)/section_size)
+            sequence_sections = [sequence[i*section_size:(i+1)*section_size] for i in range(0, sections)]
+            for index, seq in enumerate(sequence_sections):
+                nucleotide_blast_result = NCBIWWW.qblast('blastn', database, seq)
+                result = nucleotide_blast_result.read()
+                self.nucleotide_blast_results.append(result)
+                
+                # export this modular portion
+                export_path = self._paths('nucleotide_blast_results.xml', self.export_path)
+                self.paths['nucleotide_blast_results'].append(export_path)
+                with open(self.paths['nucleotide_blast_results'][-1], 'w') as nucleotide_data:
+                    nucleotide_data.write(result)
+                
+                print(f'Section {index+1}/{sequence_sections} is completed:\t{datetime.datetime.now()}')
             
         # remove the parceled information and export a new combined file
         self.nucleotide_blast_results = '\n\n'.join(self.nucleotide_blast_results)
@@ -384,6 +451,8 @@ class Codons():
             
         with open(self.paths['nucleotide_blast_results'][0], 'w') as nucleotide_data:
             nucleotide_data.write(self.nucleotide_blast_results)
+            
+        return NCBIXML.read(self.nucleotide_blast_results)
                 
     def export(self, export_name = None, export_directory = None):
         # define the simulation_path
@@ -392,11 +461,12 @@ class Codons():
             os.mkdir(self.export_path)
         
         # export the genetic and protein sequences
-        self.paths['genetic_sequence'] = os.path.join(self.export_path, 'genetic_sequence.fasta')
-        with open(self.paths['genetic_sequence'], 'w') as genes:
-            genes.write(self.gene_fasta)
+        if self.gene_fasta:
+            self.paths['genetic_sequence'] = os.path.join(self.export_path, 'genetic_sequence.fasta')
+            with open(self.paths['genetic_sequence'], 'w') as genes:
+                genes.write(self.gene_fasta)
             
-        if self.proteins != {}:
+        if self.genes != {}:
             self.paths['protein_sequence'] = os.path.join(self.export_path, 'protein_sequence.fasta')
             with open(self.paths['protein_sequence'], 'w') as proteins:
                 proteins.write(self.protein_fasta)
